import sys
import os
from typing import Callable, Dict, Any, List, Tuple
import configparser

sys.path.append('..')
import dearpygui.dearpygui as dpg
from base.msg_queue import get_msg_queue, msg, get_coroutine
from gui.lens_designer import LensDesignerWidget

CMR_CONFIG_FILE_PATH = r''
# CMR_FONT_FILE_PATH = r'C:\Windows\Fonts\msyh.ttc'
PROJECT_DIR = os.path.dirname(os.path.dirname(os.path.dirname(__file__)))


msgqueue = get_msg_queue()
coroutine = get_coroutine()

class App:
    def __init__(self) -> None:
        self._setup_init()
        self._setup_uuid()
        self._setup_style()
        self._setup_window()
        self._setup_viewport()

    def _on_app_close(self, s,a,u):
        dpg.delete_item(s)
        self._app_config.write(open(CMR_CONFIG_FILE_PATH,'a'))
        print('_on_app_close')

    def _setup_style(self):
        with dpg.font_registry():
            # Change your font here
            # dpg.add_font(CMR_FONT_FILE_PATH, 18, default_font=True)
            pass

    def _setup_init(self):
        self._app_config = configparser.ConfigParser()
        self._app_config.read(CMR_CONFIG_FILE_PATH)

    def _setup_uuid(self):
        self._gui_id_app:int = dpg.generate_uuid()
        self._lense_designer_widget:LensDesignerWidget = None

    def _gui_viewport_resize_event(self, sender, a, u):
        """
        Keep the root widget fill up the viewport
        """
        dpg.set_item_height(self._gui_id_app, a[3])
        dpg.set_item_width(self._gui_id_app, a[2])

    def _setup_viewport(self):
        if not dpg.is_viewport_created():
            icon = PROJECT_DIR+'/icon.png'
            vp = dpg.create_viewport(small_icon=icon,title='Cameray', large_icon=icon,width=1920,height=1080)
            dpg.set_viewport_resize_callback(lambda a, b:self._gui_viewport_resize_event(a, b, self._gui_id_app))
            dpg.setup_dearpygui(viewport=vp)
            dpg.show_viewport(vp)
            dpg.set_viewport_title(title='Cameray')
            # dpg.set_viewport_decorated(False)
            dpg.set_viewport_resizable(False)

    def _setup_window(self):

        with dpg.window(label="Cameray",id=self._gui_id_app,
                        on_close=self._on_app_close,
                        pos=(0, 0),
                        no_title_bar=True,
                        no_move=True,
                        no_resize=True):

            self._lense_designer_widget:LensDesignerWidget = LensDesignerWidget(self._gui_id_app)

    def _window_resize_callback(self, s,a,u):
        pass

    def show(self):
        global coroutine
        while(dpg.is_dearpygui_running()):
            while not msgqueue.empty():
                event = msgqueue.get()
                if callable(event):
                    cr = event()
                    if not cr:
                        continue
                    try:
                        next(cr)
                    except StopIteration:
                        continue
                coroutine.append(cr)

            if coroutine:
                copy_cr = coroutine
                coroutine = []
                news = []
                for cr in copy_cr:
                    try:
                        cr.send(0)
                        news.append(cr)
                    except StopIteration:
                        continue
                coroutine.extend(news)

            dpg.render_dearpygui_frame()
        dpg.cleanup_dearpygui()


def main():
    App().show()


if __name__ == '__main__':
    main()


import math
import dearpygui.dearpygui as dpg
import taichi as ti

from . import lens_preset
from typing import List, Any, Callable, Dict
from base.msg_queue import msg
from gui.widget import Widget, PropertyWidget, AttributeValueType
import numpy as np
import networkx as nx

from core.renderer import real_cam, color_buffer, taichi_render

class LensCanvasWidget(Widget):
    def __init__(self, *, parent: int, film_height=24.0, callback:Callable[[None],None]=None):
        super().__init__(parent=parent, callback=callback)
        height = 400
        with dpg.child(parent=parent,autosize_x=True,no_scrollbar=True,height=400) as self._widget_id:
            with dpg.drawlist(label='lenses',parent=self._widget_id) as self._drawlist_id:
                self.film_height = film_height
                self.axis_y = int(height / 2.0)
                self.origin_z = 0
                self.scale = 5.0
                self.lense_length = 0.0
                self.max_radius = 0.0
                self.world_matrix = np.array([[1.0,0.0,0.0], [0.0,1.0 ,0.0],[0.0,0.0,1.0]])
                self.screen_matrix = np.array([[1.0,0.0,0.0], [0.0,1.0 ,0.0],[0.0,0.0,1.0]])

        # self._drawlist_id = self.widget()

    def drawlist(self):
        return self._drawlist_id

    def _setup_transform(self,scale: float, origin_z:float, axis_y:float):
        """
        Updates transform matrix whenever the lense size or canvas size changes
        """
        self.world_matrix = np.array([
            [scale,0.0,0.0],
            [0.0,scale,0.0],
            [0.0,0.0,1.0]
            ])

        self.screen_matrix = np.array([
            [-1, -0, origin_z],
            [0, -1, axis_y],
            [0,0,1]
        ])

        self.world_to_screen = self.screen_matrix @ self.world_matrix

    def _world_to_screen(self, point: List[float]):
        return (self.world_to_screen @ np.array([point[0], point[1], 1.0]))[0:2]

    def _draw_frame(self, padding = 5):
        # rect = dpg.get_item_rect_size(self.drawlist())
        rect = (dpg.get_item_width(self.drawlist()), dpg.get_item_height(self.drawlist()))
        poly = [
            [padding,padding], # topleft
            [rect[0] - padding, padding], # topright
            [rect[0]- padding, rect[1] - padding], # bottomright
            [padding,rect[1] - padding] # bottomleft
        ]
        dpg.draw_polyline(poly, parent=self.drawlist(),closed=True)

    def _update_canvas(self, lense_length:float, lense_radius:float):
        """
        Updates whenever the lense size or canvas size changes
        """
        w, h = dpg.get_item_width(self.drawlist()), dpg.get_item_height(self.drawlist())
        rect = (w, h)
        self.scale = min(rect[0] / lense_radius, rect[1] / lense_length)
        self.axis_y = rect[1] / 2.0
        self.origin_z = rect[0]/2.0 + lense_length * self.scale / 2.0
        self._setup_transform(self.scale, self.origin_z, self.axis_y)

    def _draw_impl(self, lenses, lense_length, lense_radius):
        self._update_canvas(lense_length, lense_radius)
        self.clear_drawinglist()
        self._draw_frame()
        z = lense_length
        # draw lense groups
        for i in range(len(lenses)):
            is_stop = lenses[i][0] == 0.0
            r = lenses[i][3]/2.0
            if is_stop:
                self._draw_aperture_stop(z, r, color=[255.0, 255.0, 0.0], thickness=4.0)
            else:
                a, b = self._draw_arch(z, lenses[i][0], min(min(6.0, lenses[i][3]), abs(lenses[i][0])))
                if i > 0 and lenses[i - 1][2] != 1 and lenses[i-1][2] != 0:
                    self._draw_line(a, first)
                    self._draw_line(b, last)  # draw connection between element surface
                first, last = a, b

            z -= lenses[i][1]

        self._draw_axis()
        self._draw_film()

    def draw_lenses(self, lenses:np.ndarray):

        # workaround
        rect = dpg.get_item_rect_size(self.widget())
        dpg.configure_item(self.drawlist(), width=rect[0] - 20,height=rect[1] - 20)
        ####

        length = 0.0
        max_radius = 0.0
        for i in range(len(lenses)):
            thickness = lenses[i][1]
            if math.isnan(thickness):
                continue
            else:
                length += lenses[i][1]
            max_radius = max(max_radius, lenses[i][3] / 2.0)

        if length == 0.0 or max_radius == 0.0:
            return

        self.lense_length = length
        self.max_radius = max_radius


        self._draw_impl(lenses, length, max_radius)

    def draw_bound_rays(self, rays:np.array, count:int, color=[255.0,255.0,255.0,255.0], thickness=1.0):
        points = []
        point1 = []
        for i in range(count):
            p = rays[0][i]
            p0 = self._world_to_screen([-p[2],p[0]])
            p1 = self._world_to_screen([-p[2],-p[0]])
            points.append(p0)
            point1.append(p1)

        dpg.draw_polyline(points, parent=self.drawlist(), color=color, thickness=thickness)
        dpg.draw_polyline(point1, parent=self.drawlist(), color=color, thickness=thickness)

    def _draw_line(self, p0:List[float], p1:List[float], color=[255.0,255.0,255.0], thickness=1.0):
        dpg.draw_line(self._world_to_screen(p0), self._world_to_screen(p1), color=color, thickness=thickness,parent=self.drawlist())

    def _draw_axis(self):
        p0 = self._world_to_screen([self.lense_length + 10,0])
        p1 = self._world_to_screen([-10,0])
        dpg.draw_line(p0, p1, parent=self.drawlist())

    def _draw_film(self, color=[255.0,255.0,255.0], thickness = 4.0):
        p0 = self._world_to_screen([0, self.film_height / 2])
        p1 = self._world_to_screen([0, -self.film_height/2])
        dpg.draw_line(p0, p1, color=color, thickness=thickness, parent=self.drawlist())

    def _draw_aperture_stop(self, z:float, aperture_radius:float, color=[255.0,255.0,255.0], thickness=2.0):
        p0 = self._world_to_screen([z, aperture_radius + 10])
        p1 = self._world_to_screen([z, aperture_radius])
        p2 = self._world_to_screen([z, -aperture_radius])
        p3 = self._world_to_screen([z, -aperture_radius - 10])
        dpg.draw_line(p0, p1, color=color, thickness=thickness, parent=self.drawlist())
        dpg.draw_line(p2, p3, color=color, thickness=thickness, parent=self.drawlist())

    def _draw_arch(self, z: float, curvature_radius:float, aperture_radius:float, color=[255.0,255.0,255.0], thickness=1.0):
        """
        There is no built-in arch drawing API in DearPyGui. So arch is done with
        segmented by polylines.

        Returns the two end points of the arch
        """
        center = z - curvature_radius
        if abs(curvature_radius) < 1e-5:
            return [z, 0.0], [z, 0.0]
        half = math.asin(aperture_radius/curvature_radius)
        min_theta = -2 * half
        max_theta = 2 * half
        seg_count = 30
        points = []
        first = []
        last = []
        for i in range(seg_count):
            t = i * 1.0 / seg_count
            theta = min_theta * (1.0 - t) + t * max_theta
            p0 = (center + curvature_radius * math.cos(theta))
            p1 = curvature_radius * math.sin(theta)
            if i == 0:
                first = [p0, p1]
            elif i == seg_count - 1:
                last = [p0, p1]
            points.append(self._world_to_screen([p0, p1]))
        dpg.draw_polyline(points=points,parent=self.drawlist(),color=color,thickness=thickness)
        return first, last

    def clear_drawinglist(self):
        dpg.delete_item(self.drawlist(), children_only=True)


class LensSurface(Widget):
    def __init__(self, parent: int, callback:Callable[[None],None]=None):
        super().__init__(parent=parent,callback=callback)

class LensSphereSurface(LensSurface):
    curvature_radius = PropertyWidget('Curvature Radius', AttributeValueType.ATTRI_FLOAT,-1000,1100.0,100)
    thickness = PropertyWidget('Thickness', AttributeValueType.ATTRI_FLOAT,0.0,1000.0,100)
    eta = PropertyWidget('Eta', AttributeValueType.ATTRI_FLOAT,0.0,100.0,100)
    aperture_radius = PropertyWidget('Aperture Radius', AttributeValueType.ATTRI_FLOAT,0.0,1000.0,100)
    def __init__(self, parent: int, callback:Callable[[None],None]):
        super().__init__(parent=parent,callback=callback)
        with dpg.tree_node(label='SphereElement',parent=parent) as self._widget_id:
            self.curvature_radius = 0.0
            self.thickness = 0.0
            self.eta = 0.0
            self.aperture_radius = 0.0

    def dump(self):
        return [ self.curvature_radius,self.thickness,self.eta,self.aperture_radius]

    def property_changed(self, s, a, u):
        self._invoke_update()

    def load(self, data:List[float]= [0.0,0.0,0.0,0.0]):
        self.curvature_radius = data[0]
        self.thickness = data[1]
        self.eta = data[2]
        self.aperture_radius = data[3]



class WidgetNode(Widget):
    def __init__(self,*, name:str, parent:int, callback:Callable[[Any],Any]=None):
        super(WidgetNode, self).__init__(parent=parent, callback=callback)
        self._attri_dict:Dict[str,Any] = {}
        self._callback = callback
        self._input_attr = None
        self._output_attr = None
        with dpg.node(label=name,parent=parent, user_data=self) as self._widget_id:
            pass

    def add_attribute(self, attri_name:str, attri_type:int):
        if attri_name not in self._attri_dict.keys():
            with dpg.node_attribute(label=attri_name, attribute_type=attri_type, parent=self.widget(),user_data=self.widget()) as attri:
                self._attri_dict[attri_name] = (attri, {})
                return attri
        return None

    def get_attribute(self, attri_name:str):
        return self._attri_dict.get(attri_name, (None, {}))[0]

    def remove_attribute(self, attri_name):
        if attri_name in self._attri_dict.keys():
            dpg.delete_item(self._attri_dict[attri_name][0])

    def add_value(self,*,attri_name:str,
                        value_name:str,
                        value_type:int,
                        default_value:Any,
                        size:int=4,
                        callback:Callable[[Any], Any]=None):

        attri_id = self.get_attribute(attri_name)
        attri_id, value_dict = self._attri_dict.get(attri_name, (None, {}))
        if attri_id is None:
            print('No corresponding attribute :', attri_name)
            return

        width = 100

        if value_name in value_dict.keys():
            print(value_name, 'has already existed in attribute ', attri_name)
            return None
        else:
            if value_type == AttributeValueType.ATTRI_FLOAT:
                value_id = dpg.add_input_float(label=value_name, callback=callback,default_value=default_value,parent=attri_id,width=width)
            elif value_type == AttributeValueType.ATTRI_FLOATX:
                value_id = dpg.add_input_floatx(label=value_name, callback=callback,default_value=default_value,size=size, parent=attri_id,width=width)
            elif value_type == AttributeValueType.ATTRI_INT:
                value_id = dpg.add_input_int(label=value_name,callback=callback,default_value=default_value, parent=attri_id,width=width)
            value_dict[value_name] = value_id

        return value_id


    def get_attri_value_item_id(self,attri_name:str, value_name:str):
        return self._attri_dict.get(attri_name, (-1, {}))[1].get(value_name, None)

    def get_value(self, attri_name:str, value_name:str):
        item_id = self.get_attri_value_item_id(attri_name=attri_name,value_name=value_name)
        if item_id is not None:
            return dpg.get_value(item_id)

        print('No such value: ', value_name, ' of ', attri_name)
        return None

    def set_value(self, attri_name:str, value_name:str, value:Any):
        item_id = self.get_attri_value_item_id(attri_name=attri_name,value_name=value_name)
        if item_id is not None:
            dpg.configure_item(item=item_id, value=Any)
            return
        print('No such value: ', value_name, ' of ', attri_name)


class MidNode(WidgetNode):
    def __init__(self, *, name: str, parent: int, callback: Callable[[Any], Any]):
        super().__init__(name=name, parent=parent, callback=callback)
        self._input_end = self.add_attribute('input', dpg.mvNode_Attr_Input)
        self._output_end = self.add_attribute('output', dpg.mvNode_Attr_Output)

    def input_end(self):
        return self._input_end

    def output_end(self):
        return self._output_end


class OutputNode(WidgetNode):
    def __init__(self, *, name: str, parent: int, callback: Callable[[Any], Any]):
        super().__init__(name=name, parent=parent, callback=callback)
        self._output_end = self.add_attribute('output', dpg.mvNode_Attr_Output)

    def output_end(self):
        return self._output_end

class InputNode(WidgetNode):
    def __init__(self, *, name: str, parent: int, callback: Callable[[Any], Any]):
        super().__init__(name=name, parent=parent, callback=callback)
        self._input_end = self.add_attribute('input', dpg.mvNode_Attr_Input)

    def input_end(self):
        return self._input_end

class SceneNodeParam(Widget):
    focus_depth = PropertyWidget(name='FocusDepth', property_type=AttributeValueType.ATTRI_FLOAT,min_value=0.0, max_value=10000.0,width=100)
    enable_focus = PropertyWidget(name='EnableFocus', property_type=AttributeValueType.ATTRI_BOOL, width=100)
    def __init__(self, *, parent: int, callback: Callable[[Any], None]):
        super().__init__(parent=parent, callback=callback)
        self._widget_id = parent
        self.focus_depth = 1.0
        self.enable_focus = True

class SceneNode(OutputNode):
    def __init__(self,parent:int, value_update_callback:Callable[[Any], None]=None):
        super().__init__(name='Scene',parent=parent,callback=value_update_callback)
        attri = self.add_attribute(attri_name='focus depth',attri_type=dpg.mvNode_Attr_Static)
        self._param = SceneNodeParam(parent=attri,callback=self.callback())

    def get_focus_depth(self):
        return self._param.focus_depth

    def get_focus_state(self):
        return self._param.enable_focus

class FilmNodeParam(Widget):
    film_size = PropertyWidget(name='FilmSize', property_type=AttributeValueType.ATTRI_FLOATX,min_value=0.0,max_value=1000.0,size=2,width=100)
    render_window = PropertyWidget(name='RenderWindow', property_type=AttributeValueType.ATTRI_BOOL, width=100)
    def __init__(self, *, parent: int, callback: Callable[[Any], None]):
        super().__init__(parent=parent, callback=callback)
        self._widget_id = parent
        self.film_size = (36.00,24.00)
        self.render_window = True
        self._image = ImageViewer(parent=self.widget())
        self._enable_render = True

        if self._enable_render:
            self._render()


    @msg
    def _render(self):
        i = 0
        color_buffer.from_numpy(np.zeros((800, 600, 3)))
        last_t = 0.0
        while True:
            taichi_render()
            interval = 50
            if i % interval == 0 and i > 0:
                img = color_buffer.to_numpy() * (1 / (i + 1))
                img = np.sqrt(img / img.mean() * 0.24)
                # var = np.var(img)
                # img = img * 255.0
                # Image.fromarray(img.astype('uint8')).convert('RGB').save('output.jpg')
                self._image.from_numpy(img)

            i+=1
            if self._enable_render:
                yield
            else:
                break
        return

    def property_changed(self, s, a, u):
        '''
        Filters the RenderWindow property changes
        '''
        if s == getattr(self, '_RenderWindow'):
            dpg.configure_item(self._image.widget(), show=a)
            self._enable_render = a
            if self._enable_render:
                self._render()
        return super().property_changed(s, a, u)


class ImageViewer(Widget):
    def __init__(self, parent):
        super().__init__(parent=parent)
        with dpg.child(parent=parent, label='Image',height=600,width=800) as self._widget_id:
            self._texture_container = dpg.add_texture_registry(label='Texture')
            self._width:int = 0
            self._height:int = 0
            self._texture_id:int = None

    def _release_texture(self):
        if self._texture_id is not None:
            dpg.delete_item(self._texture_id)
        self._texture_id = None

    def _set_or_recreate_texture(self, width, height, norm_rbga:List[float]):
        if width <= 0 or height <= 0:
            return
        if not self.valid() or self._width != width or self._height != height:
            self._release_texture()
            rect = dpg.get_item_rect_size(self.parent())
            self._texture_id = dpg.add_dynamic_texture(width, height, norm_rbga, parent=self._texture_container)
            dpg.add_image(self._texture_id,parent=self.widget(), width=rect[0], height=rect[1])
            self._width = width
            self._height = height
            return
        dpg.set_value(self._texture_id, norm_rbga)

    def set_image_norm_rgba(self,width:int, height:int, rgba:List[float]):
        """
        each channel of rgba is range from [0 1]
        """
        self._set_or_recreate_texture(width,height,rgba)

    def from_numpy(self, data:np.ndarray):
        shape = data.shape
        norm_rgba = []
        if shape[2] > 4:
            return
        if shape[2] == 3:
            rgba = np.concatenate((data, np.ones((shape[0],shape[1], 1),dtype=np.float32)), axis=2)
            self.set_image_norm_rgba(shape[1],shape[0],rgba.flatten())
        elif shape[2] == 4:
            self.set_image_norm_rgba(shape[1],shape[0],data.flatten())
        elif shape[2] == 2 or shape[2] == 1:
            pass

    def to_numpy(self)->np.ndarray:
        raise NotImplementedError

    def width(self)->int:
        return self._width

    def height(self)->int:
        return self._height

    def valid(self)->bool:
        return self._width > 0 and self._height > 0 and self._texture_id != None


class FilmNode(InputNode):
    def __init__(self, parent:int,value_update_callback:Callable[[Any], None]=None):
        super().__init__(name='Film',parent=parent,callback=value_update_callback)
        attri = self.add_attribute(attri_name='film size',attri_type=dpg.mvNode_Attr_Static)
        self._param = FilmNodeParam(parent=attri,callback=self.callback())

    def get_film_size(self):
        return self._param.film_size

    def get_keep_rendering(self):
        return self._param.render_window

class ApertureSurface(LensSurface):
    thickness = PropertyWidget('Thickness', AttributeValueType.ATTRI_FLOAT,0.0,100.0,100)
    aperture_radius = PropertyWidget('Aperture Radius', AttributeValueType.ATTRI_FLOAT,0.0,100.0,100)
    def __init__(self, parent: int, callback:Callable[[None],None]):
        super().__init__(parent=parent,callback=callback)
        self._widget_id = parent
        self.thickness = 0.0
        self.aperture_radius = 0.0

    def dump(self):
        return [ 0.0,self.thickness, 0.0 ,self.aperture_radius]

    def load(self, data:List[float]= [0.0,0.0,0.0,0.0]):
        self.thickness = data[1]
        self.aperture_radius = data[3]

class ApertureStop(MidNode):
    def __init__(self, *, parent: int,value_update_callback:Callable[[Any], None]=None):
        super().__init__(name='Aperture Stop', parent=parent, callback=value_update_callback)
        self.static_attri = self.add_attribute('Aperture', attri_type=dpg.mvNode_Attr_Static)
        self.aperture_surface = ApertureSurface(self.static_attri, self.callback())

    def get_surface(self, ind:int):
        return self.aperture_surface

    def get_surface_count(self):
        return 1

class LensSurfaceGroup(MidNode):
    def __init__(self,*,name:str,parent:int, update_callback:Callable[[Any], Any]=None):
        super().__init__(name=name, parent=parent,callback=update_callback)
        self._lense_surface_group:List[LensSurface] = []
        self._surface_data_value_id:List[int] = []
        self.input_attri_item_id = self.add_attribute('surface count', attri_type=dpg.mvNode_Attr_Static)

        self.count_attri = self.add_value(attri_name='surface count',
        value_name="Surface Count",
        value_type=AttributeValueType.ATTRI_INT,
        default_value=0,
        callback=lambda s,a,u:self._update_surface(int(a)))


    def _update_surface(self, count):
        """
        """
        cur_count = len(self._lense_surface_group)
        print('surface count changed: ',count, cur_count)

        delta = count - cur_count
        if delta > 0:
            for _ in range(delta):
                surf = LensSphereSurface(self.input_attri_item_id, callback=self.callback())
                self._lense_surface_group.append(surf)
        elif delta < 0:
            for item in self._lense_surface_group[count:]:
                item.delete()
            del self._lense_surface_group[count:]

        self._invoke_update()

    def load(self, raw_group_data: List[List[float]]):
        self._clear_surface_data()
        surfs = []
        for s in raw_group_data:
            surf = LensSphereSurface(self.input_attri_item_id, callback=self.callback())
            surf.load(s)
            surfs.append(surf)

        self._lense_surface_group = surfs
        dpg.set_value(self.count_attri, len(self._lense_surface_group))
        self._invoke_update()

    def get_surface(self, ind:int):
        return self._lense_surface_group[ind]

    def get_surface_count(self):
        return len(self._lense_surface_group)

    def _clear_surface_data(self):
        for surf in self._lense_surface_group:
            surf.delete()
        self._lense_surface_group = []

    def clear_surface_data(self):
        """
        Invode update signal
        """
        self._clear_surface_data()
        self._invoke_update()

    @staticmethod
    def create_sphere_lense_group(*, name, parent, group_data:List[List[float]], callback=None):
        group = LensSurfaceGroup(name=name, parent=parent,update_callback=callback)
        group.block_callback(True)
        group.load(group_data)
        group.block_callback(False)
        return group


class EditorEventType:
    EVENT_NODE_ADD= 0x000
    EVENT_NODE_DELETE = 0x001
    EVENT_NODE_UPDATE = 0x002
    EVENT_NODE_DELETE_ALL = 0x003

    EVENT_LINK_ADD = 0x200
    EVENT_LINK_DELETE = 0x201


    EVENT_SURFACE_ADD = 0x300
    EVENT_SURFACE_DELETE = 0x301
    EVENT_SURFACE_ATTRIBUTE_CHANGED = 0x302

    EVENT_UPDATE_ALL = 0xFFF


class ToolBar(Widget):
    def __init__(self, *, parent: int, callback: Callable[[Any], None]):
        super().__init__(parent=parent, callback=callback)
        self.add_node_button = dpg.add_button(label='Add Node')
        dpg.add_same_line()
        self.remove_node_button = dpg.add_button(label='Remove Nodes')
        dpg.add_same_line()
        self.remove_link_button = dpg.add_button(label='Remove Links')
        dpg.add_same_line()
        self.clear_all_button = dpg.add_button(label='Clear')
        dpg.add_same_line()
        self.save_button = dpg.add_button(label='Save')
        dpg.add_same_line()
        self.open_button = dpg.add_button(label='Open')
        dpg.add_same_line()
        self.preset_combo = dpg.add_combo(list(lens_preset.lens_data.keys()),label='Lens Preset',width=150)
        dpg.add_same_line()
        self.auto_arrange = dpg.add_button(label='Auto Arrange')


class LensEditorWidget(Widget):
    def __init__(self,*,update_callback:Callable[[Any],None], parent: int):
        super().__init__(parent=parent, callback=update_callback)
        self._lense_data:List[Dict[str, List[float]]]= []
        self._valid_lenses = False
        self._editor_id = -1

        with dpg.group(horizontal=False,parent=parent) as self._widget_id:
            self._toolbar = ToolBar(parent=self._widget_id,callback=self.callback())
            with dpg.node_editor(parent=self._widget_id,callback=self._link_add_callback, delink_callback=self._link_delete_callback, height = 800) as self._editor_id:
                pass
            dpg.configure_item(self._toolbar.add_node_button,callback = lambda s,a,u:self.add_lens_group(LensSurfaceGroup(name='LensGroup', parent=self._editor_id, update_callback=self.callback())))
            dpg.configure_item(self._toolbar.clear_all_button,callback = lambda s,a,u:self.clear_lense_group())
            dpg.configure_item(self._toolbar.preset_combo,callback = lambda s,a,u:self.set_lense_data(lens_preset.lens_data.get(a,[])))
            dpg.configure_item(self._toolbar.remove_node_button,callback = lambda s,a,u:self.remove_selected_nodes())
            dpg.configure_item(self._toolbar.remove_link_button,callback = lambda s,a,u:self.remove_selected_links())
            dpg.configure_item(self._toolbar.auto_arrange,callback = lambda s,a,u:self.auto_arrange())

        # self._editor_id = self.widget()
        self.widget_g = nx.Graph()
        self.attri_g = nx.Graph()

        self._add_default_node()

    def _link_add_callback(self, sender:int, app_data:Any, user_data:Any):
        """
        input_end
        output_end
        """
        self._add_link_impl(app_data[0], app_data[1])
        self._invoke_update(event=EditorEventType.EVENT_LINK_ADD)

    def _link_delete_callback(self, sender:int, app_data:Any, user_data:Any):
        """
        app_data: edge item
        """
        self._remove_link_impl(app_data)
        self._invoke_update(event=EditorEventType.EVENT_LINK_DELETE)

    def _add_link_impl(self, output_end, input_end):
        output_node, input_node = dpg.get_item_user_data(output_end), dpg.get_item_user_data(input_end)
        deleted_attri_edge = []
        deleted_node_edge = []
        for nbr, datadict in self.attri_g.adj[input_end].items():
            deleted_attri_edge.append((nbr, input_end))
            edge_item = datadict.get('edge_item', None)
            if edge_item:
                dpg.delete_item(edge_item)
                adj_output_node = dpg.get_item_user_data(nbr)
                deleted_node_edge.append((adj_output_node, input_node))

        self.attri_g.remove_edges_from(deleted_attri_edge)
        self.widget_g.remove_edges_from(deleted_node_edge)

        link = dpg.add_node_link(output_end, input_end, parent=self._editor_id, user_data={'input_node':input_node,'output_node':output_node,'input_end':input_end,'output_end':output_end})
        self.widget_g.add_edge(output_node, input_node, edge_item=link)
        self.attri_g.add_edge(output_end, input_end, edge_item=link)

    def _remove_link_impl(self, edge_item):
        udata = dpg.get_item_user_data(edge_item)
        input_node, output_node = udata['input_node'], udata['output_node']
        self.widget_g.remove_edge(output_node,input_node)
        self.attri_g.remove_edge(udata['output_end'],udata['input_end'])
        dpg.delete_item(edge_item)

    def _add_node_impl(self, lens_group_node):
        self.widget_g.add_node(lens_group_node.widget())
        if isinstance(lens_group_node, SceneNode):
            self.attri_g.add_node(lens_group_node.output_end())
        elif isinstance(lens_group_node, FilmNode):
            self.attri_g.add_node(lens_group_node.input_end())
        else:
            self.attri_g.add_node(lens_group_node.input_end())
            self.attri_g.add_node(lens_group_node.output_end())

    def _remove_selected_nodes_impl(self):
        """
        TODO::

        dpg.get_selected_nodes may hava a potential BUG

        """
        print("_remove_selected_nodes_impl::dpg.get_selected_nodes may hava a potential BUG")
        selected_nodes = dpg.get_selected_nodes(self._editor_id)
        for x in selected_nodes:
            self._remove_node_impl(dpg.get_item_user_data(x))

    def _remove_selected_links_impl(self):
        selected_links = dpg.get_selected_links(self._editor_id)
        for link in selected_links:
            self._remove_link_impl(link)

    def _remove_node_impl(self, lens_group_node):
        item = lens_group_node.widget()
        self.widget_g.remove_node(item)
        attri_ends = []
        if isinstance(lens_group_node, SceneNode):
            attri_ends.append(lens_group_node.output_end())
        elif isinstance(lens_group_node, FilmNode):
            attri_ends.append(lens_group_node.input_end())
        else:
            attri_ends.append(lens_group_node.input_end())
            attri_ends.append(lens_group_node.output_end())

        deleted_edge = []
        for end in attri_ends:
            for nbr, datadict in self.attri_g.adj[end].items():
                edge_item = datadict['edge_item']
                deleted_edge.append((nbr, end))
                dpg.delete_item(edge_item)

        self.attri_g.remove_edges_from(deleted_edge)
        lens_group_node.delete()

    def _remove_all_node_impl(self):
        list(map(lambda x: self._remove_node_impl(dpg.get_item_user_data(x)), list(self.widget_g.nodes)))

    def _add_default_node(self):
        self._scene_node = SceneNode(self._editor_id, self.callback())
        self._film_node = FilmNode(self._editor_id, self.callback())
        self._add_node_impl(self._scene_node)
        self._add_node_impl(self._film_node)

    def add_lens_group(self, node:LensSurfaceGroup):
        self._add_node_impl(node)
        self._invoke_update(event=EditorEventType.EVENT_NODE_ADD)

    def remove_lense_group(self, nw:Widget):
        self._remove_node_impl(nw)
        self._invoke_update(event=EditorEventType.EVENT_NODE_DELETE)

    def remove_selected_nodes(self):
        self._remove_selected_nodes_impl()
        self._invoke_update(event=EditorEventType.EVENT_NODE_DELETE)

    def remove_selected_links(self):
        self._remove_selected_links_impl()
        self._invoke_update(event=EditorEventType.EVENT_LINK_DELETE)

    def clear_lense_group(self):
        self._remove_all_node_impl()
        self._add_default_node()
        self._invoke_update(event=EditorEventType.EVENT_NODE_DELETE)

    def auto_arrange(self):
        rect = dpg.get_item_rect_size(self._editor_id)
        all_nodes = self.get_lense_group()
        s = len(all_nodes)
        if s < 2:
            return

        s = int(s / 2.0)
        w_interval = rect[0] / (s + 2)
        h_interval = rect[1] / 2.0
        for i in range(len(all_nodes)):
            y = (i % 2) * h_interval + 100
            x = (i / 2) * w_interval + 100
            dpg.configure_item(all_nodes[i].widget(),pos=(x,y))


    def get_lense_group(self):
        res = list(nx.all_simple_paths(self.widget_g, self._scene_node.widget(), self._film_node.widget()))
        ret = []
        if len(res) > 0:
            ret = list(map(lambda x: dpg.get_item_user_data(x), res[0]))
        return ret

    def get_lense_data(self):
        groups = self.get_lense_group()
        lense_data = []
        for g in range(1, len(groups) - 1):  # excludes scene and film node
            c = groups[g].get_surface_count()
            for ind in range(c):
                surf = groups[g].get_surface(ind)
                lense_data.append(surf.dump())
        return lense_data

    def set_lense_data(self, lense_data:List[List[float]]):
        self._remove_all_node_impl()
        self._add_default_node()

        # parse lense data
        stack = []
        output_end = self._scene_node.output_end()
        for surf in lense_data:
            if surf[2] != 1.0 and surf[0] != 0:  # surface begin
                stack.append(surf)
            elif surf[2] == 1.0 and surf[0] != 0: # surface end
                stack.append(surf)
                lense_group = LensSurfaceGroup.create_sphere_lense_group(name='LensGroup',parent=self._editor_id,group_data=stack, callback=self.callback())
                self._add_node_impl(lense_group)
                input_end = lense_group.input_end()
                self._add_link_impl(output_end=output_end,input_end=input_end)
                output_end = lense_group.output_end()
                stack = []
            elif surf[0] == 0.0 and surf[2] == 0.0:  # aperture surface
                aperture = ApertureStop(parent=self._editor_id,value_update_callback=self.callback())
                aperture.block_callback(True)
                aperture.get_surface(0).load(surf)
                self._add_node_impl(aperture)
                input_end = aperture.input_end()
                self._add_link_impl(output_end=output_end,input_end=input_end)
                output_end = aperture.output_end()
                aperture.block_callback(False)

        if len(stack) > 0:
            print('Wrong lense data')

        self._add_link_impl(output_end=output_end, input_end=self._film_node.input_end())
        # closed
        # parse data end

        self.auto_arrange()
        self._invoke_update(event=EditorEventType.EVENT_NODE_UPDATE)

    def get_film_size(self):
        return self._film_node.get_film_size()

    def get_focus_depth(self):
        return self._scene_node.get_focus_depth()

    def get_focus_state(self):
        return self._scene_node.get_focus_state()

    def get_keep_rendering(self):
        return self._film_node.get_keep_rendering()




class LensDesignerWidget(Widget):

    def __init__(self, parent: int):
        super().__init__(parent=parent)
        with dpg.group(parent=parent,horizontal=False) as self._widget_id:
            self._lense_canvas: LensCanvasWidget = LensCanvasWidget(parent=self.widget(),callback=self._canvas_update)
            self._node_editor: LensEditorWidget = LensEditorWidget(update_callback=self._editor_update, parent=self.widget())

            pos = [0.0, 3.0, 24.0]
            center = [0.0, 0.0, 0.0]
            world_up = [0.0, 1.0, 0.0]
            # self.camera = RealisticCamera(pos, center, world_up)
            self._paint_canvas()


    def _editor_update(self, *args, **kwargs):
        """
        Update Realistic camera here
        """
        lense_data = self._node_editor.get_lense_data()
        self._canvas_update(lense_data = lense_data)

    def _canvas_update(self, *args, **kwargs):
        self._update_camera(kwargs['lense_data'])

    def _paint_canvas(self):
        ray_points = real_cam.get_ray_points()
        new_lense_data = real_cam.get_lenses_data()
        self._lense_canvas.draw_lenses(np.array(new_lense_data))
        self._lense_canvas.draw_bound_rays(ray_points, real_cam.get_element_count() + 2, color=[0, 0, 255])


    @msg
    def _update_camera(self, lense_data):
        real_cam.load_lens_data(lense_data)
        self._node_editor.get_focus_state() and real_cam.refocus(self._node_editor.get_focus_depth())
        if self._node_editor.get_keep_rendering():
            # real_cam.recompute_exit_pupil()
            color_buffer.from_numpy(np.zeros((800, 600, 3)))
        real_cam.gen_draw_rays_from_film()
        self._paint_canvas()


import queue
import functools

_msg_queue = queue.Queue(50)

_main_loop_coroutine = []

def msg(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        global _msg_queue
        _msg_queue.put(lambda:func(*args,**kwargs))
    return wrapper


def get_msg_queue():
    return _msg_queue

def get_coroutine():
    return _main_loop_coroutine



import dearpygui.dearpygui as dpg
from dearpygui.logger import mvLogger
from typing import Callable, Any

_logger = None

def init_global_logger(parent:int):
    global _logger
    if _logger is None:
        _logger = mvLogger(parent)

def get_logger()->mvLogger:
    return _logger


def _config(sender, keyword, user_data):
    widget_type = dpg.get_item_type(sender)
    items = user_data

    if widget_type == "mvAppItemType::mvRadioButton":
        value = True

    else:
        keyword = dpg.get_item_label(sender)
        value = dpg.get_value(sender)

    if isinstance(user_data, list):
        for item in items:
            dpg.configure_item(item, **{keyword: value})
    else:
        dpg.configure_item(items, **{keyword: value})

class Widget:
    def __init__(self,*, parent:int,callback:Callable[[Any],None]=None):
        self._widget_id:int = None
        self._parent_id:int = parent
        self._block = False
        self._callback = callback

    def widget(self)->int:
        return self._widget_id

    def parent(self)->int:
        return self._parent_id

    def __hash__(self):
        return self._widget_id

    def property_changed(self, s, a, u):
        self._invoke_update(sender=s, app_data=a, user_data=u)

    def block_callback(self, block):
        self._block = block

    def _invoke_update(self, *args, **kwargs):
        not self._block and callable(self._callback) and self._callback(*args, **kwargs)

    def callback(self):
        return self._callback

    def delete(self):
        self._widget_id and dpg.delete_item(self._widget_id)
        self._widget_id = None

    def __del__(self):
        self.delete()

class AttributeValueType:
    ATTRI_FLOAT = 0
    ATTRI_FLOATX = 1
    ATTRI_INT = 2
    ATTRI_BOOL = 3

def PropertyWidget(name:str, property_type:int, min_value:Any=None, max_value:Any=None, width=20,height=10,size=4):
    storage_name = '_' + name
    @property
    def prop(self:Widget):
        return dpg.get_value(getattr(self, storage_name))

    @prop.setter
    def prop(self:Widget, value):
        if not hasattr(self, storage_name):
            if property_type == AttributeValueType.ATTRI_FLOAT:
                item_id = dpg.add_input_float(
                    label=name,
                    default_value=value,
                    min_value=min_value,
                    max_value=max_value,
                    width=width,
                    parent=self.widget(),
                    callback=self.property_changed)
            elif property_type == AttributeValueType.ATTRI_FLOATX:
                item_id = dpg.add_input_floatx(
                    label=name,
                    default_value=value,
                    min_value=min_value,
                    max_value=max_value,
                    width=width,
                    parent=self.widget(),
                    size=size,
                    callback=self.property_changed)
            elif property_type == AttributeValueType.ATTRI_INT:
                item_id = dpg.add_input_int(
                    label=name,
                    default_value=value,
                    min_value=min_value,
                    max_value=max_value,
                    width=width,
                    parent=self.widget(),
                    callback=self.property_changed)
            elif property_type == AttributeValueType.ATTRI_BOOL:
                item_id = dpg.add_checkbox(label=name,parent=self.widget(),default_value=value,callback=self.property_changed)
            setattr(self, storage_name, item_id)
        else:
            item_id = getattr(self, storage_name)
            dpg.set_value(item_id, value)
    return prop



import taichi as ti
import math

eps = 1e-4
inf = 1e10


@ti.func
def out_dir(n):
    u = ti.Vector([1.0, 0.0, 0.0])
    if ti.abs(n[1]) < 1 - 1e-3:
        u = n.cross(ti.Vector([0.0, 1.0, 0.0])).normalized()
    v = n.cross(u)
    phi = 2 * math.pi * ti.random(ti.f32)
    r = ti.random(ti.f32)
    ay = ti.sqrt(r)
    ax = ti.sqrt(1 - r)
    return ax * (ti.cos(phi) * u + ti.sin(phi) * v) + ay * n


@ti.func
def reflect(d, n):
    # Assuming |d| and |n| are normalized
    return d - 2.0 * d.dot(n) * n


@ti.func
def refract(d, n, ni_over_nt):
    # Assuming |d| and |n| are normalized
    has_r, rd = 0, d
    dt = d.dot(n)
    discr = 1.0 - ni_over_nt * ni_over_nt * (1.0 - dt * dt)
    if discr > 0.0:
        has_r = 1
        rd = (ni_over_nt * (d - n * dt) - n * ti.sqrt(discr)).normalized()
    else:
        rd *= 0.0
    return has_r, rd


@ti.func
def ray_aabb_intersection(box_min, box_max, o, d):
    intersect = 1

    near_int = -inf
    far_int = inf

    for i in ti.static(range(3)):
        if d[i] == 0:
            if o[i] < box_min[i] or o[i] > box_max[i]:
                intersect = 0
        else:
            i1 = (box_min[i] - o[i]) / d[i]
            i2 = (box_max[i] - o[i]) / d[i]

            new_far_int = ti.max(i1, i2)
            new_near_int = ti.min(i1, i2)

            far_int = ti.min(new_far_int, far_int)
            near_int = ti.max(new_near_int, near_int)

    if near_int > far_int:
        intersect = 0
    return intersect, near_int, far_int


# (T + x d)(T + x d) = r * r
# T*T + 2Td x + x^2 = r * r
# x^2 + 2Td x + (T * T - r * r) = 0

refine = True


@ti.func
def intersect_sphere(pos, d, center, radius):
    T = pos - center
    A = 1.0
    B = 2.0 * T.dot(d)
    C = T.dot(T) - radius * radius
    delta = B * B - 4.0 * A * C
    dist = inf
    hit_pos = ti.Vector([0.0, 0.0, 0.0])

    if delta > -1e-4:
        delta = ti.max(delta, 0)
        sdelta = ti.sqrt(delta)
        ratio = 0.5 / A
        ret1 = ratio * (-B - sdelta)
        dist = ret1
        if ti.static(refine):
            if dist < inf:
                # refinement
                old_dist = dist
                new_pos = pos + d * dist
                T = new_pos - center
                A = 1.0
                B = 2.0 * T.dot(d)
                C = T.dot(T) - radius * radius
                delta = B * B - 4 * A * C
                if delta > 0:
                    sdelta = ti.sqrt(delta)
                    ratio = 0.5 / A
                    ret1 = ratio * (-B - sdelta) + old_dist
                    if ret1 > 0:
                        dist = ret1
                        hit_pos = new_pos + ratio * (-B - sdelta) * d
                    else:
                        pass
                        #ret2 = ratio * (-B + sdelta) + old_dist
                        #if ret2 > 0:
                        #  dist = ret2
                        #  hit_pos = new_pos + ratio * (-B + sdelta) * d
                else:
                    dist = inf

    return dist, hit_pos


@ti.func
def ray_plane_intersect(pos, d, pt_on_plane, norm):
    dist = inf
    hit_pos = ti.Vector([0.0, 0.0, 0.0])
    denom = d.dot(norm)
    if abs(denom) > eps:
        dist = norm.dot(pt_on_plane - pos) / denom
        hit_pos = pos + d * dist
    return dist, hit_pos


@ti.func
def point_aabb_distance2(box_min, box_max, o):
    p = ti.Vector([0.0, 0.0, 0.0])
    for i in ti.static(range(3)):
        p[i] = ti.max(ti.min(o[i], box_max[i]), box_min[i])
    return (p - o).norm_sqr()


@ti.func
def sphere_aabb_intersect(box_min, box_max, o, radius):
    return point_aabb_distance2(box_min, box_max, o) < radius * radius


@ti.func
def sphere_aabb_intersect_motion(box_min, box_max, o1, o2, radius):
    lo = 0.0
    hi = 1.0
    while lo + 1e-5 < hi:
        m1 = 2 * lo / 3 + hi / 3
        m2 = lo / 3 + 2 * hi / 3
        d1 = point_aabb_distance2(box_min, box_max, (1 - m1) * o1 + m1 * o2)
        d2 = point_aabb_distance2(box_min, box_max, (1 - m2) * o1 + m2 * o2)
        if d2 > d1:
            hi = m2
        else:
            lo = m1

    return point_aabb_distance2(box_min, box_max,
                                (1 - lo) * o1 + lo * o2) < radius * radius


@ti.func
def inside(p, c, r):
    return (p - c).norm_sqr() <= r * r


@ti.func
def inside_left(p, c, r):
    return inside(p, c, r) and p[0] < c[0]


@ti.func
def inside_right(p, c, r):
    return inside(p, c, r) and p[0] > c[0]


def Vector2(x, y):
    return ti.Vector([x, y])


@ti.func
def inside_taichi(p_):
    p = p_
    ret = -1
    if not inside(p, Vector2(0.50, 0.50), 0.52):
        if ret == -1:
            ret = 0
    if not inside(p, Vector2(0.50, 0.50), 0.495):
        if ret == -1:
            ret = 1
    p = Vector2(0.5, 0.5) + (p - Vector2(0.5, 0.5))
    if inside(p, Vector2(0.50, 0.25), 0.08):
        if ret == -1:
            ret = 1
    if inside(p, Vector2(0.50, 0.75), 0.08):
        if ret == -1:
            ret = 0
    if inside(p, Vector2(0.50, 0.25), 0.25):
        if ret == -1:
            ret = 0
    if inside(p, Vector2(0.50, 0.75), 0.25):
        if ret == -1:
            ret = 1
    if p[0] < 0.5:
        if ret == -1:
            ret = 1
    else:
        if ret == -1:
            ret = 0
    return ret



import sys
from numpy.lib.type_check import real
import taichi as ti
import time
import math
import numpy as np
from .renderer_utils import ray_aabb_intersection, intersect_sphere, ray_plane_intersect, reflect, refract
from .realistic import RealisticCamera

ti.init(arch=ti.gpu)
res = (600, 800)
color_buffer = ti.Vector.field(3, dtype=ti.f32, shape=res)
# color_buffer = ti.Vector.field(3, dtype=ti.f32)

# ti.root.dense(ti.ji,res).place(color_buffer)

count_var = ti.field(ti.i32, shape=(1, ))

max_ray_depth = 10
eps = 1e-4
inf = 1e10
fov = 0.8

camera_pos = ti.Vector([0.0, 1.8, 10.0])

mat_none = 0
mat_lambertian = 1
mat_specular = 2
mat_glass = 3
mat_light = 4

light_y_pos = 30.0 - eps
light_x_min_pos = -0.25
light_x_range = 5
light_z_min_pos = 100.0
light_z_range = 12
light_area = light_x_range * light_z_range
light_min_pos = ti.Vector([light_x_min_pos, light_y_pos, light_z_min_pos])
light_max_pos = ti.Vector([
    light_x_min_pos + light_x_range, light_y_pos,
    light_z_min_pos + light_z_range
])
light_color = ti.Vector(list(np.array([0.9, 0.85, 0.7])))
light_normal = ti.Vector([0.0, -1.0, 0.0])

# No absorbtion, integrates over a unit hemisphere
lambertian_brdf = 1.0 / math.pi
# diamond!
refr_idx = 2.4

# right near sphere
sp1 = [0.0, 2.0, 225.0]
sp1_center = ti.Vector(sp1)
sp1_radius = 2.0
# left far sphere
sp2_center = ti.Vector([-0.28, 0.55, 0.8])
sp2_radius = 0.32



pos = [0.0, 3.0, 240.0]   # mm
center = [0.0,0.0,0.0]
world_up = [0.0, 1.0, 0.0]
real_cam = RealisticCamera(pos, center, world_up)

def make_box_transform_matrices():
    rad = math.pi / 10.0
    c, s = math.cos(rad), math.sin(rad)
    rot = np.array([[c, 0, s, 0], [0, 1, 0, 0], [-s, 0, c, 0], [0, 0, 0, 1]])
    translate = np.array([
        [1, 0, 0, -0.7],
        [0, 1, 0, 0],
        [0, 0, 1, 0.7],
        [0, 0, 0, 1],
    ])
    m = translate @ rot
    m_inv = np.linalg.inv(m)
    m_inv_t = np.transpose(m_inv)
    return ti.Matrix(m_inv), ti.Matrix(m_inv_t)


# left box
box_min = ti.Vector([-4.0, -1.0, 15.0])
box_max = ti.Vector([4.0, 30.0, 25.00])
box_m_inv, box_m_inv_t = make_box_transform_matrices()


@ti.func
def intersect_light(pos, d, tmax):
    hit, t, _ = ray_aabb_intersection(light_min_pos, light_max_pos, pos, d)
    if hit and 0 < t < tmax:
        hit = 1
    else:
        hit = 0
        t = inf
    return hit, t


@ti.func
def ray_aabb_intersection2(box_min, box_max, o, d):
    # Compared to ray_aabb_intersection2(), this also returns the normal of
    # the nearest t.
    intersect = 1

    near_t = -inf
    far_t = inf
    near_face = 0
    near_is_max = 0

    for i in ti.static(range(3)):
        if d[i] == 0:
            if o[i] < box_min[i] or o[i] > box_max[i]:
                intersect = 0
        else:
            i1 = (box_min[i] - o[i]) / d[i]
            i2 = (box_max[i] - o[i]) / d[i]

            new_far_t = max(i1, i2)
            new_near_t = min(i1, i2)
            new_near_is_max = i2 < i1

            far_t = min(new_far_t, far_t)
            if new_near_t > near_t:
                near_t = new_near_t
                near_face = int(i)
                near_is_max = new_near_is_max

    near_norm = ti.Vector([0.0, 0.0, 0.0])
    if near_t > far_t:
        intersect = 0
    if intersect:
        # TODO: Issue#1004...
        if near_face == 0:
            near_norm[0] = -1 + near_is_max * 2
        elif near_face == 1:
            near_norm[1] = -1 + near_is_max * 2
        else:
            near_norm[2] = -1 + near_is_max * 2

    return intersect, near_t, far_t, near_norm


@ti.func
def mat_mul_point(m, p):
    hp = ti.Vector([p[0], p[1], p[2], 1.0])
    hp = m @ hp
    hp /= hp[3]
    return ti.Vector([hp[0], hp[1], hp[2]])


@ti.func
def mat_mul_vec(m, v):
    hv = ti.Vector([v[0], v[1], v[2], 0.0])
    hv = m @ hv
    return ti.Vector([hv[0], hv[1], hv[2]])


@ti.func
def ray_aabb_intersection2_transformed(box_min, box_max, o, d):
    # Transform the ray to the box's local space
    obj_o = mat_mul_point(box_m_inv, o)
    obj_d = mat_mul_vec(box_m_inv, d)
    intersect, near_t, _, near_norm = ray_aabb_intersection2(
        box_min, box_max, obj_o, obj_d)
    if intersect and 0 < near_t:
        # Transform the normal in the box's local space to world space
        near_norm = mat_mul_vec(box_m_inv_t, near_norm)
    else:
        intersect = 0
    return intersect, near_t, near_norm


@ti.func
def intersect_scene(pos, ray_dir):
    closest, normal = inf, ti.Vector.zero(ti.f32, 3)
    c, mat = ti.Vector.zero(ti.f32, 3), mat_none

    # right near sphere
    cur_dist, hit_pos = intersect_sphere(pos, ray_dir, sp1_center, sp1_radius)
    if 0 < cur_dist < closest:
        closest = cur_dist
        normal = (hit_pos - sp1_center).normalized()
        c, mat = ti.Vector([1.0, 1.0, 1.0]), mat_glass
    # left box
    hit, cur_dist, pnorm = ray_aabb_intersection2_transformed(
        box_min, box_max, pos, ray_dir)
    if hit and 0 < cur_dist < closest:
        closest = cur_dist
        normal = pnorm
        c, mat = ti.Vector([0.8, 0.5, 0.4]), mat_specular

    # left
    pnorm = ti.Vector([1.0, 0.0, 0.0])
    cur_dist, _ = ray_plane_intersect(pos, ray_dir, ti.Vector([-40.0, 0.0,
                                                               0.0]), pnorm)
    if 0 < cur_dist < closest:
        closest = cur_dist
        normal = pnorm
        c, mat = ti.Vector([0.65, 0.05, 0.05]), mat_lambertian
    # right
    pnorm = ti.Vector([-1.0, 0.0, 0.0])
    cur_dist, _ = ray_plane_intersect(pos, ray_dir, ti.Vector([40.0, 0.0, 0.0]),
                                      pnorm)
    if 0 < cur_dist < closest:
        closest = cur_dist
        normal = pnorm
        c, mat = ti.Vector([0.12, 0.45, 0.15]), mat_lambertian
    # bottom
    gray = ti.Vector([0.93, 0.93, 0.93])
    pnorm = ti.Vector([0.0, 1.0, 0.0])
    cur_dist, _ = ray_plane_intersect(pos, ray_dir, ti.Vector([0.0, -1.0, 0.0]),
                                      pnorm)
    if 0 < cur_dist < closest:
        closest = cur_dist
        normal = pnorm
        c, mat = gray, mat_lambertian
    # top
    pnorm = ti.Vector([0.0, -1.0, 0.0])
    cur_dist, _ = ray_plane_intersect(pos, ray_dir, ti.Vector([0.0, 30.0, 0.0]),
                                      pnorm)
    if 0 < cur_dist < closest:
        closest = cur_dist
        normal = pnorm
        c, mat = gray, mat_lambertian
    # far
    pnorm = ti.Vector([0.0, 0.0, 1.0])
    cur_dist, _ = ray_plane_intersect(pos, ray_dir, ti.Vector([0.0, 0.0, 0.0]),
                                      pnorm)
    if 0 < cur_dist < closest:
        closest = cur_dist
        normal = pnorm
        c, mat = gray, mat_lambertian
    # light
    hit_l, cur_dist = intersect_light(pos, ray_dir, closest)
    if hit_l and 0 < cur_dist < closest:
        # technically speaking, no need to check the second term
        closest = cur_dist
        normal = light_normal
        c, mat = light_color, mat_light

    return closest, normal, c, mat


@ti.func
def visible_to_light(pos, ray_dir):
    a, b, c, mat = intersect_scene(pos, ray_dir)
    return mat == mat_light


@ti.func
def dot_or_zero(n, l):
    return max(0.0, n.dot(l))


@ti.func
def mis_power_heuristic(pf, pg):
    # Assume 1 sample for each distribution
    f = pf**2
    g = pg**2
    return f / (f + g)


@ti.func
def compute_area_light_pdf(pos, ray_dir):
    hit_l, t = intersect_light(pos, ray_dir, inf)
    pdf = 0.0
    if hit_l:
        l_cos = light_normal.dot(-ray_dir)
        if l_cos > eps:
            tmp = ray_dir * t
            dist_sqr = tmp.dot(tmp)
            pdf = dist_sqr / (light_area * l_cos)
    return pdf


@ti.func
def compute_brdf_pdf(normal, sample_dir):
    return dot_or_zero(normal, sample_dir) / math.pi


@ti.func
def sample_area_light(hit_pos, pos_normal):
    # sampling inside the light area
    x = ti.random() * light_x_range + light_x_min_pos
    z = ti.random() * light_z_range + light_z_min_pos
    on_light_pos = ti.Vector([x, light_y_pos, z])
    return (on_light_pos - hit_pos).normalized()


@ti.func
def sample_brdf(normal):
    # cosine hemisphere sampling
    # first, uniformly sample on a disk (r, theta)
    r, theta = 0.0, 0.0
    sx = ti.random() * 2.0 - 1.0
    sy = ti.random() * 2.0 - 1.0
    if sx >= -sy:
        if sx > sy:
            # first region
            r = sx
            div = abs(sy / r)
            if sy > 0.0:
                theta = div
            else:
                theta = 7.0 + div
        else:
            # second region
            r = sy
            div = abs(sx / r)
            if sx > 0.0:
                theta = 1.0 + sx / r
            else:
                theta = 2.0 + sx / r
    else:
        if sx <= sy:
            # third region
            r = -sx
            div = abs(sy / r)
            if sy > 0.0:
                theta = 3.0 + div
            else:
                theta = 4.0 + div
        else:
            # fourth region
            r = -sy
            div = abs(sx / r)
            if sx < 0.0:
                theta = 5.0 + div
            else:
                theta = 6.0 + div
    # Malley's method
    u = ti.Vector([1.0, 0.0, 0.0])
    if abs(normal[1]) < 1 - eps:
        u = normal.cross(ti.Vector([0.0, 1.0, 0.0]))
    v = normal.cross(u)

    theta = theta * math.pi * 0.25
    costt, sintt = ti.cos(theta), ti.sin(theta)
    xy = (u * costt + v * sintt) * r
    zlen = ti.sqrt(max(0.0, 1.0 - xy.dot(xy)))
    return xy + zlen * normal


@ti.func
def sample_direct_light(hit_pos, hit_normal, hit_color):
    direct_li = ti.Vector([0.0, 0.0, 0.0])
    fl = lambertian_brdf * hit_color * light_color
    light_pdf, brdf_pdf = 0.0, 0.0
    # sample area light
    to_light_dir = sample_area_light(hit_pos, hit_normal)
    if to_light_dir.dot(hit_normal) > 0:
        light_pdf = compute_area_light_pdf(hit_pos, to_light_dir)
        brdf_pdf = compute_brdf_pdf(hit_normal, to_light_dir)
        if light_pdf > 0 and brdf_pdf > 0:
            l_visible = visible_to_light(hit_pos, to_light_dir)
            if l_visible:
                w = mis_power_heuristic(light_pdf, brdf_pdf)
                nl = dot_or_zero(to_light_dir, hit_normal)
                direct_li += fl * w * nl / light_pdf
    # sample brdf
    brdf_dir = sample_brdf(hit_normal)
    brdf_pdf = compute_brdf_pdf(hit_normal, brdf_dir)
    if brdf_pdf > 0:
        light_pdf = compute_area_light_pdf(hit_pos, brdf_dir)
        if light_pdf > 0:
            l_visible = visible_to_light(hit_pos, brdf_dir)
            if l_visible:
                w = mis_power_heuristic(brdf_pdf, light_pdf)
                nl = dot_or_zero(brdf_dir, hit_normal)
                direct_li += fl * w * nl / brdf_pdf
    return direct_li


@ti.func
def schlick(cos, eta):
    r0 = (1.0 - eta) / (1.0 + eta)
    r0 = r0 * r0
    return r0 + (1 - r0) * ((1.0 - cos)**5)


@ti.func
def sample_ray_dir(indir, normal, hit_pos, mat):
    u = ti.Vector([0.0, 0.0, 0.0])
    pdf = 1.0
    if mat == mat_lambertian:
        u = sample_brdf(normal)
        pdf = max(eps, compute_brdf_pdf(normal, u))
    elif mat == mat_specular:
        u = reflect(indir, normal)
    elif mat == mat_glass:
        cos = indir.dot(normal)
        ni_over_nt = refr_idx
        outn = normal
        if cos > 0.0:
            outn = -normal
            cos = refr_idx * cos
        else:
            ni_over_nt = 1.0 / refr_idx
            cos = -cos
        has_refr, refr_dir = refract(indir, outn, ni_over_nt)
        refl_prob = 1.0
        if has_refr:
            refl_prob = schlick(cos, refr_idx)
        if ti.random() < refl_prob:
            u = reflect(indir, normal)
        else:
            u = refr_dir
    return u.normalized(), pdf


stratify_res = 5
inv_stratify = 1.0 / 5.0


@ti.kernel
def taichi_render():
    for u, v in color_buffer:
        weight, r = real_cam.gen_ray_of(v, res[0]-u)
        if weight <= 0.0:
            continue
        ray_dir = r[1]
        pos = r[0]

        acc_color = ti.Vector([0.0, 0.0, 0.0])
        throughput = ti.Vector([1.0, 1.0, 1.0])
        depth = 0
        while depth < max_ray_depth:
            closest, hit_normal, hit_color, mat = intersect_scene(pos, ray_dir)
            if mat == mat_none:
                break

            hit_pos = pos + closest * ray_dir
            hit_light = (mat == mat_light)
            if hit_light:
                acc_color += throughput * light_color
                break
            elif mat == mat_lambertian:
                acc_color += throughput * sample_direct_light(
                    hit_pos, hit_normal, hit_color)

            depth += 1
            ray_dir, pdf = sample_ray_dir(ray_dir, hit_normal, hit_pos, mat)
            pos = hit_pos + 1e-4 * ray_dir
            if mat == mat_lambertian:
                throughput *= lambertian_brdf * hit_color * dot_or_zero(
                    hit_normal, ray_dir) / pdf
            else:
                throughput *= hit_color
        color_buffer[u, v] += weight*acc_color


# gui = ti.GUI('Realistic camera', res)
last_t = time.time()
i = 0

real_cam.refocus(4.5)
# real_cam.refocus(np.linalg.norm(np.array(sp1) - real_cam.get_position()))
real_cam.recompute_exit_pupil()


class Renderer:
    def __init__(self):
        self.camera = None
        self._iter = 0

    def render(self):
        taichi_render()

    def clear(self):
        color_buffer.from_numpy(np.zeros(res))
        self._iter = 0

    def var(self):
        pass

    def refocus(self, depth):
        global real_cam
        real_cam.refocus(depth)

    def recompute_exit_pupil(self):
        global real_cam
        real_cam.recompute_exit_pupil()

    def get_camera(self):
        global real_cam
        return real_cam

    def get_color_buffer_to_numpy(self):
        return color_buffer.to_numpy()


# while gui.running:
#     taichi_render()
#     interval = 2000
#     if i % interval == 0 and i > 0:
#         img = color_buffer.to_numpy() * (1 / (i + 1))
#         img = np.sqrt(img / img.mean() * 0.24)
#         var = np.var(img)
#         if var < 0.11790:
#             ti.imwrite(img, 'output.png')
#             break
#         print("{:.2f} samples/s ({} iters, var={})".format(
#             interval / (time.time() - last_t), i, var))
#         last_t = time.time()
#         gui.set_image(img)
#         gui.show()
#     i += 1



import math
import taichi as ti
import numpy as np
from typing import List, Dict
from .renderer_utils import refract

max_elements = 30
max_draw_rays = 20
pupil_interval_count = 64
eps = 1e-5
inf = 9999999.0

def convert_raw_data_from_dict(raw_data:List[List[float]]):
    """
    Args: raw_data
    for example:
    [
        # curvature radius, thickness, index of refraction, aperture diameter
        [29.475,3.76,1.67,25.2],
        [84.83,0.12,1,25.2],
        [19.275,4.025,1.67,23],
        [40.77,3.275,1.699,23],
        [12.75,5.705,1,18],
        [0,4.5,0,17.1],
        [-14.495,1.18,1.603,17],
        [40.77,6.065,1.658,20],
        [-20.385,0.19,1,20],
        [437.065,3.22,1.717,20],
        [-39.73,5.0,1,20]
    ]
    """
    dict_data = []
    for elem in raw_data:
        dict_data.append(
            {'curvature_radius':elem[0],
            'thickness':elem[1],
            'eta':elem[2],
            'aperture_diameter':elem[3]
            })
    return dict_data

def convert_dict_data_from_raw(dict_data:List[Dict[str, List[float]]]):
    raw_data = []
    for elem_dict in dict_data:
        raw_data.append([
            elem_dict.get('curvature_radius', 0),
            elem_dict.get('thickness', 0),
            elem_dict.get('eta', 0),
            elem_dict.get('aperture_diameter', 0)
            ])
    return raw_data

@ti.func
def lerp(val, begin ,end):
    return begin * (1.0 - val) + val * end

@ti.func
def bound_union_with(bmin, bmax, pos):
    return ti.min(bmin, pos), ti.max(bmax, pos)

@ti.func
def make_bound2():
    return ti.Vector([inf, inf]), ti.Vector([-inf, -inf])

@ti.func
def inside_aabb(bmin, bmax, pos):
    return all(bmin <= pos) and all(pos <= bmax)



@ti.data_oriented
class RealisticCamera:
    def __init__(self, camera_pos, center, world_up):
        self.vignet = ti.field(ti.i32, shape=())

        self.curvature_radius = ti.field(ti.f32)
        self.thickness = ti.field(ti.f32)
        self.eta = ti.field(ti.f32)
        self.aperture_radius = ti.field(ti.f32)
        self.exitPupilBoundMin = ti.Vector.field(2, dtype=ti.f32)
        self.exitPupilBoundMax = ti.Vector.field(2, dtype=ti.f32)
        self.draw_rays = ti.Vector.field(3, dtype=ti.f32)

        ti.root.dense(ti.i, (max_elements, )).place(self.curvature_radius, self.thickness, self.eta, self.aperture_radius)
        ti.root.dense(ti.i, (pupil_interval_count, )).place(self.exitPupilBoundMin, self.exitPupilBoundMax)
        ti.root.dense(ti.ij, (max_draw_rays, max_elements + 2)).place(self.draw_rays)

        self.camera2world_point = ti.Matrix([[1.0,0.0,0.0,0.0],[0.0,1.0,0.0,0.0],[0.0,0.0,1.0,0.0],[0.0,0.0,0.0,1.0]])
        self.camera2world_vec = ti.Matrix([[1.0,0.0,0.0],[0.0,1.0,0.0],[0.0,0.0,1.0]])

        self.shutter = 0.1
        self.film_width = 36
        self.film_height = 24
        self.film_diagnal = math.sqrt(self.film_height * self.film_height + self.film_width * self.film_width)
        self.pixel_width = 800
        self.pixel_height = 600
        self.camera_pos = np.array(camera_pos)

        self._elem_count = 0
        self.load_lens_data([])
        self.set_camera(camera_pos, center, world_up)

    def get_resolution(self):
        return [self.pixel_width, self.pixel_height]

    def get_position(self):
        return self.camera_pos

    def set_camera(self, eye, center, world_up):
        """
        setup camera2world_pos transformation
        """
        eye = np.array(eye)
        center = np.array(center)
        world_up = np.array(world_up)

        self.camera_pos = eye

        def normalize(v):
            norm = np.linalg.norm(v)
            return v / norm if norm > 0.0 else np.array([0.0,0.0,0.0])

        direction = normalize(center - eye)
        right = normalize(np.cross(direction, world_up))
        up = normalize(np.cross(right, direction))

        self.camera2world_point = ti.Matrix([
            [1.0/1000.0,0.0,0.0,0.0],
            [0.0,1.0/1000.0,0.0,0.0],
            [0.0,0.0,1.0/1000.0,0.0],
            [0.0,0.0,0.0,1.0]
        ]) @ ti.Matrix([
            [right[0], up[0], direction[0], 1000.0 * eye[0]],
            [right[1], up[1], direction[1], 1000.0 * eye[1]],
            [right[2], up[2], direction[2], 1000.0 * eye[2]],
            [0.0,0.0,0.0,1.0]])

        self.camera2world_vec = ti.Matrix([
            [right[0], up[0], direction[0]],
            [right[1], up[1], direction[1]],
            [right[2], up[2], direction[2]]
            ])

    def get_element_count(self):
        return self._elem_count

    def load_lens_data(self, lenses:List[List[float]]=[[]]):
        wide22 = [
            # curvature radius, thickness, index of refraction, aperture diameter
            [35.98738, 1.21638, 1.54, 23.716],
            [11.69718, 9.9957, 1, 17.996],
            [13.08714, 5.12622, 1.772, 12.364],
            [22.63294, 1.76924, 1.617, 9.812],
            [71.05802, 0.8184, 1, 9.152],
            [0,        2.27766,0, 8.756],
            [9.58584,2.43254,1.617,8.184],
            [11.28864,0.11506,1,9.152],
            [166.7765,3.09606,1.713,10.648],
            [7.5911,1.32682,1.805,11.44],
            [16.7662,3.98068,1,12.276],
            [7.70286,1.21638,1.617,13.42],
            [11.97328,10,1,17.996]
        ]
        dgauss50 = [
            [29.475,3.76,1.67,25.2],
            [84.83,0.12,1,25.2],
            [19.275,4.025,1.67,23],
            [40.77,3.275,1.699,23],
            [12.75,5.705,1,18],
            [0,4.5,0,17.1],
            [-14.495,1.18,1.603,17],
            [40.77,6.065,1.658,20],
            [-20.385,0.19,1,20],
            [437.065,3.22,1.717,20],
            [-39.73,5.0,1,20]
        ]
        telephoto=[
            [21.851,0	,1.529,19.0],
            [-34.546,5.008,1.599,17.8],
            [108.705,1.502,1.0,16.6],
            [  0,1.127,0,16.2],
            [-12.852,26.965,1.613,12.6],
            [19.813,1.502,1.603,13.4],
            [-20.378,5.008,1.0,14.8]
        ]
        telephoto250=[
            [54.6275,12.52,1.529,47.5],
            [-86.365,3.755,1.599,44.5],
            [271.7625,2.8175,1,41.5],
            [0,67.4125,0,40.5],
            [-32.13,3.755,1.613,31.5],
            [49.5325,12.52,1.603,33.5],
            [-50.945,0,1,37]
        ]

        lenses = lenses if lenses else dgauss50
        self._elem_count = len(lenses)
        self._lenses_data = lenses.copy()
        for _ in range(max(0, max_elements - self._elem_count)):
            lenses.append([0 for j in range(4)])
        a = np.array(lenses).transpose()
        self.curvature_radius.from_numpy(a[0])
        self.thickness.from_numpy(a[1])
        self.eta.from_numpy(a[2])
        self.aperture_radius.from_numpy(a[3]/2.0)

    def get_lenses_data(self):
        a = self.curvature_radius.to_numpy()[0:self._elem_count]
        b = self.thickness.to_numpy()[0:self._elem_count]
        c = self.eta.to_numpy()[0:self._elem_count]
        d = self.aperture_radius.to_numpy()[0:self._elem_count] * 2.0
        return np.stack((a,b,c,d)).transpose()

    @ti.func
    def camera_2_world(self, o, d):
        """
        Transform the ray from camera space to world space

        o: origin of the ray
        d: direction of the ray
        """
        wo = self.camera2world_point @ ti.Vector([o.x, o.y, o.z, 1.0])
        wd = self.camera2world_vec @ d
        return ti.Vector([wo.x,wo.y,wo.z]), wd

    @ti.func
    def rear_z(self):
        return self.thickness[self._elem_count - 1]

    @ti.func
    def front_z(self):
        z = 0.0
        for i in self.thickness:
            z += self.thickness[i]
        return z

    @ti.func
    def rear_radius(self):
        return self.curvature_radius[self._elem_count - 1]

    @ti.func
    def rear_aperture(self):
        return self.aperture_radius[self._elem_count - 1]

    @ti.kernel
    def recompute_exit_pupil(self):
        """
        pre-process exit pupil of the lens system
        """

        rearZ = self.rear_z()
        if rearZ <= 0.0:
            print('Not focus')
        rearRadius = self.rear_aperture()
        samples = 1024 * 1024
        half = 2.0 * rearRadius
        proj_bmin, proj_bmax = ti.Vector([-half, -half]), ti.Vector([half, half])
        for i in range(pupil_interval_count):
            r0 = ti.cast(i, ti.f32) / pupil_interval_count * self.film_diagnal / 2.0
            r1 = ti.cast(i + 1, ti.f32) / pupil_interval_count * self.film_diagnal / 2.0
            bmin, bmax = make_bound2()
            count = 0
            for j in range(samples):
                u, v= ti.random(), ti.random()
                film_pos = ti.Vector([lerp(ti.cast(j, ti.f32)/samples, r0, r1), 0.0, 0.0])
                x, y = lerp(u, -half, half), lerp(v, -half, half)
                lens_pos = ti.Vector([x, y, rearZ])
                if inside_aabb(bmin, bmax, ti.Vector([x, y])):
                    ti.atomic_add(count, 1)
                else:
                    ok, _, _ = self.gen_ray_from_film(film_pos, (lens_pos - film_pos).normalized())
                    if ok:
                        bmin, bmax = bound_union_with(bmin,bmax, ti.Vector([x, y]))
                        ti.atomic_add(count, 1)

            if count == 0:
                bmin, bmax = proj_bmin, proj_bmax

            # extents pupil bound
            delta = 2 * (proj_bmax - proj_bmin).norm() / ti.sqrt(samples)
            bmin -= delta
            bmax += delta

            self.exitPupilBoundMin[i] = bmin
            self.exitPupilBoundMax[i] = bmax

    @ti.func
    def sample_exit_pupil(self, film_pos, uv):
        """
        filme_pos: sampled point on the film
        uv: 2d sample

        Returns the sample point in the lenses space and the pupil area
        """

        r = film_pos.norm()
        index = ti.cast(ti.min(r / self.film_diagnal * 2.0 * pupil_interval_count, pupil_interval_count - 1), ti.i32)
        bmin, bmax = self.exitPupilBoundMin[index], self.exitPupilBoundMax[index]
        area = (bmax - bmin).dot(ti.Vector([1.0, 1.0]))
        sampled = lerp(uv, bmin, bmax)
        sint = film_pos.y / r if abs(r) >= eps else 0.0
        cost = film_pos.x / r if abs(r) >= eps else 1.0
        return ti.Vector(
            [
                cost * sampled.x - sint * sampled.y,
                sint * sampled.x + cost * sampled.y,
                self.rear_z()
            ]), area

    @ti.func
    def gen_ray(self, film_uv, lens_uv):

        """
        film_uv: samples on film
        lens_uv: samples on lens

        Returns:
            weight: non-zero if exit ray exists otherwise returns 0
            (ro, rd): exit ray
            film_pos: sampled point on film
        """
        extent = ti.Vector([self.film_width, self.film_height])
        film_pos_xy = lerp(film_uv, ti.Vector([0.0, 0.0]), extent)

        film_pos_xy = ti.Vector([film_pos_xy.x - self.film_width/2.0, self.film_height /2.0 - film_pos_xy.y])
        lens_pos, area = self.sample_exit_pupil(film_pos_xy, lens_uv)
        film_pos = ti.Vector([film_pos_xy.x, film_pos_xy.y, 0.0])
        o, d = film_pos, lens_pos - film_pos
        exit, out_o ,out_d = self.gen_ray_from_film(o, d)

        weight = 0.0
        if not exit:
            self.vignet[None] += 1

        if exit:
            cost = out_d.z
            cos4t = cost * cost * cost * cost
            weight = self.shutter * cos4t * area /(self.rear_z() * self.rear_z())
        pos = lerp(film_uv, ti.Vector([0.0, 0.0]), ti.Vector([self.pixel_width, self.pixel_height]))
        return weight, self.camera_2_world(out_o, out_d), pos

    @ti.func
    def gen_ray_of(self, px, py):
        """
        px, py: pixel position of final image
        """
        lens_uv = ti.Vector([ti.random(), ti.random()])
        u, v = ti.cast(px, ti.f32) / self.pixel_width, ti.cast(py, ti.f32) / self.pixel_height
        film_uv = ti.Vector([u, v])

        extent = ti.Vector([self.film_width, self.film_height])
        film_pos_xy = lerp(film_uv, ti.Vector([0.0, 0.0]), extent)

        film_pos_xy = ti.Vector([film_pos_xy.x - self.film_width/2.0, self.film_height /2.0 - film_pos_xy.y])
        lens_pos, area = self.sample_exit_pupil(film_pos_xy, lens_uv)
        film_pos = ti.Vector([film_pos_xy.x, film_pos_xy.y, 0.0])
        o, d = film_pos, lens_pos - film_pos
        exit, out_o ,out_d = self.gen_ray_from_film(o, d)

        weight = 0.0
        if not exit:
            self.vignet[None] += 1

        if exit:
            cost = out_d.z
            cos4t = cost * cost * cost * cost
            weight = self.shutter * cos4t * area /(self.rear_z() * self.rear_z())

        return weight, self.camera_2_world(out_o, out_d)

    @ti.func
    def compute_cardinal_points(self, in_ro, out_ro, out_rd):
        """
        Returns the z coordinate of the principal plane the the focal point
        (fz, pz) in the lenses space
        note: input vectors are in camera space
        """
        tf = -out_ro.x / out_rd.x
        tp = (in_ro.x - out_ro.x) / out_rd.x
        return -(out_ro + out_rd * tf).z, -(out_ro + out_rd * tp).z


    @ti.func
    def compute_thick_lens_approximation(self):
        """
        Returns the focal length and the z of principal plane
        return fz1, pz1, fz2, pz2 in lense space
        """

        x = self.film_diagnal * 0.001
        so = ti.Vector([x, 0.0, self.front_z() + 1.0])
        sd = ti.Vector([0.0, 0.0, -1.0])
        fo = ti.Vector([x, 0.0, self.rear_z() - 1.0])
        fd = ti.Vector([0.0, 0.0, 1.0])
        ok1, o1, d1 = self.gen_ray_from_scene(so, sd)
        ok2, o2, d2 = self.gen_ray_from_film(fo, fd)
        assert ok1 == True and ok2 == True
        fz, pz = self.compute_cardinal_points(so, o1, d1)
        fz1, pz1 = self.compute_cardinal_points(fo, o2, d2)
        assert fz1 < pz1 and pz < fz
        return fz, pz, fz1, pz1

    @ti.func
    def get_focal_length(self):
        fz, pz ,fz1 ,pz1 = self.compute_thick_lens_approximation()
        return fz - pz, pz1 - fz1

    @ti.func
    def focus_thick_camera(self, focus_distance):
        """
        focus_distance > 0
        """
        fz1, pz1, fz2, pz2 = self.compute_thick_lens_approximation()
        f = fz1 - pz1
        assert f > 0
        z = -focus_distance
        delta = 0.5 * (pz2 - z + pz1 - ti.sqrt((pz2 - z - pz1)*(pz2-z-4*f-pz1) ))
        return self.thickness[self._elem_count - 1] + delta

    @ti.kernel
    def refocus(self, focus_distance:ti.f32):
        rf = self.focus_thick_camera(focus_distance * 1000.0)
        self.thickness[self._elem_count - 1] = rf

    @ti.func
    def intersect_with_sphere(self,
                            center,
                            radius,
                            ro,
                            rd):
        """
        center: z depth of lens sphere center
        """

        o = ro - ti.Vector([0.0, 0.0, center])
        ok = True

        # translate the sphere to original
        A = rd.x * rd.x + rd.y * rd.y + rd.z * rd.z
        B = 2 * ( rd.x * o.x + rd.y * o.y + rd.z * o.z)
        C = o.x * o.x + o.y * o.y + o.z * o.z - radius * radius
        delta = B * B - 4 * A * C
        if delta < 0:
            ok = False

        root_delta = ti.sqrt(delta)
        t = 0.0
        n = ti.Vector([0.0,0.0,0.0])

        if ok:
            q = 0.0
            if B < 0:
                q = -0.5 * (B - root_delta)
            else:
                q = -0.5 * (B + root_delta)
            t0, t1 = q/A , C/q
            if t0 > t1:
                t0, t1 = t1, t0

            # t0, t1 = (-B - root_delta) / (2 * A), (-B + root_delta) / (2 * A)

            closer = (rd.z > 0) ^ (radius < 0)
            t = ti.min(t0, t1) if closer else ti.max(t0, t1)
            if t < 0:
                ok = False
            if ok:
                n = (o + t * rd).normalized()
                n = -n if n.dot(-rd) < 0.0 else n

        return ok, t, n


    @ti.func
    def gen_ray_from_scene(self, ori, dir):
        ro, rd = ti.Vector([ti.cast(ori.x,ti.f32), ori.y, -ori.z]), ti.Vector([ti.cast(dir.x,ti.f32), dir.y, -dir.z]).normalized()
        elemZ = -self.front_z()
        ok = True
        t = 0.0
        n = ti.Vector([0.0,0.0,0.0])
        for _ in range(1):  # force the inner loop serialized so that break could be used
            for i in range(self._elem_count):
                is_stop = self.curvature_radius[i] == 0.0
                if is_stop:
                    t = (elemZ - ro.z) / rd.z
                else:
                    radius = self.curvature_radius[i]
                    centerZ = elemZ + radius
                    isect, t, n = self.intersect_with_sphere(centerZ, radius, ro, rd)
                    if not isect:
                        ok = False
                        break

                assert t > 0.0

                hit = ro + rd * t
                r = hit.x * hit.x + hit.y * hit.y
                if r > self.aperture_radius[i] * self.aperture_radius[i]:  # out of the element aperture
                    ok = False
                    break
                ro = ti.Vector([hit.x, hit.y, hit.z])

                if not is_stop:
                    # refracted by lens
                    etaI = 1.0 if i == 0 or self.eta[i - 1] == 0.0 else self.eta[i - 1]
                    etaT = self.eta[i] if self.eta[i] != 0.0 else 1.0
                    rd.normalized()
                    has_r, d = refract(rd, n, etaI/etaT)
                    if not has_r:
                        ok = False
                        break
                    rd = ti.Vector([d.x, d.y, d.z])

                elemZ += self.thickness[i]

        return ok, ti.Vector([ro.x, ro.y, -ro.z]), ti.Vector([rd.x, rd.y, -rd.z]).normalized()

    @ti.func
    def gen_ray_from_film(self, ori, dir):
        """
        Input ray is the initial ray sampled from film to the rear lens element.
        Returns True and the output ray if the ray could be pass the lens system
        or returns False
        """
        ro, rd = ti.Vector([ori.x, ori.y, -ori.z]), ti.Vector([dir.x, dir.y, -dir.z]).normalized()
        elemZ = 0.0
        ok = True
        t = 0.0
        n = ti.Vector([0.0,0.0,0.0])
        for _ in range(1):  # force the inner loop serialized so that break could be allowed
            for ii in range(self._elem_count):
                i = self._elem_count - ii - 1
                elemZ -= self.thickness[i]
                is_stop = self.curvature_radius[i] == 0.0
                if is_stop:
                    if rd.z >= 0.0:
                        ok = False
                        break
                    t = (elemZ - ro.z) / rd.z
                else:
                    radius = self.curvature_radius[i]
                    centerZ = elemZ + radius
                    isect, t, n = self.intersect_with_sphere(centerZ, radius, ro, rd)
                    if not isect:
                        ok = False
                        break

                assert t > 0.0
                hit = ro + rd * t
                r = hit.x * hit.x + hit.y * hit.y
                if r > self.aperture_radius[i] * self.aperture_radius[i]:  # out of the element aperture
                    ok = False
                    break

                ro = ti.Vector([hit.x, hit.y, hit.z])

                if not is_stop:
                    # refracted by lens
                    etaI = self.eta[i]
                    etaT = self.eta[i - 1] if i > 0 and self.eta[i - 1] != 0.0 else 1.0   # the outer of 0-th element is air, whose eta is 1.0
                    # rd.normalized()
                    has_r, d = refract(rd, n, etaI/etaT)
                    if not has_r:
                        ok = False
                        break
                    rd = ti.Vector([d.x, d.y, d.z])

        return ok, ti.Vector([ro.x, ro.y, -ro.z]), ti.Vector([rd.x, rd.y, -rd.z]).normalized()

    @ti.kernel
    def gen_draw_rays_from_film(self):
        """
        draw the bound ray
        """
        r = self.aperture_radius[self._elem_count - 1]
        step = 0.01
        count = ti.cast(r / step,ti.i32)
        for j in range(1):
            for i in range(count):
                y = r - i * step
                ori, dir = ti.Vector([0.0, 0.0, 0.0]), ti.Vector([y, 0.0, self.rear_z()])
                ok, a, b = self.gen_ray_from_film(ori, dir)
                if ok:
                    self.draw_ray_from_film(ori, dir, 0)
                    break

    @ti.kernel
    def gen_draw_rays_from_scene(self):
        pass

    @ti.func
    def draw_ray_from_film(self, ori, dir, ind):
        ro, rd = ti.Vector([ori.x, ori.y, -ori.z]), ti.Vector([dir.x, dir.y, -dir.z]).normalized()
        elemZ = 0.0
        ok = True
        t = 0.0
        n = ti.Vector([0.0,0.0,0.0])
        for _ in range(1):  # force the inner loop serialized so that break could be allowed
            for ii in range(self._elem_count):
                i = self._elem_count - ii - 1
                elemZ -= self.thickness[i]
                is_stop = self.curvature_radius[i] == 0.0
                if is_stop:
                    if rd.z >= 0.0:
                        ok = False
                        break
                    t = (elemZ - ro.z) / rd.z
                else:
                    radius = self.curvature_radius[i]
                    centerZ = elemZ + radius
                    isect, t, n = self.intersect_with_sphere(centerZ, radius, ro, rd)
                    if not isect:
                        ok = False
                        break

                hit = ro + rd * t
                r = hit.x * hit.x + hit.y * hit.y
                if r > self.aperture_radius[i] * self.aperture_radius[i]:  # out of the element aperture
                    ok = False
                    break

                self.draw_rays[ind, ii] = ro

                ro = ti.Vector([hit.x, hit.y, hit.z])

                if not is_stop:
                    # refracted by lens
                    etaI = self.eta[i]
                    etaT = self.eta[i - 1] if i > 0 and self.eta[i - 1] != 0.0 else 1.0   # the outer of 0-th element is air, whose eta is 1.0
                    has_r, d = refract(rd, n, etaI/etaT)
                    if not has_r:
                        ok = False
                        break
                    rd = ti.Vector([d.x, d.y, d.z])

        self.draw_rays[ind, self._elem_count] = ro
        self.draw_rays[ind, self._elem_count + 1] = ro + rd * 20.0
        # return ok, ti.Vector([ro.x, ro.y, -ro.z]), ti.Vector([rd.x, rd.y, -rd.z]).normalized()

    @ti.func
    def draw_ray_from_scene(self, ori, dir, ind):
        ro, rd = ti.Vector([ti.cast(ori.x,ti.f32), ori.y, -ori.z]), ti.Vector([ti.cast(dir.x,ti.f32), dir.y, -dir.z]).normalized()
        elemZ = -self.front_z()
        ok = True
        t = 0.0
        n = ti.Vector([0.0,0.0,0.0])
        for _ in range(1):  # force the inner loop serialized so that break could be used
            for i in range(self._elem_count):
                is_stop = self.curvature_radius[i] == 0.0
                if is_stop:
                    t = (elemZ - ro.z) / rd.z
                else:
                    radius = self.curvature_radius[i]
                    centerZ = elemZ + radius
                    isect, t, n = self.intersect_with_sphere(centerZ, radius, ro, rd)
                    if not isect:
                        ok = False
                        break

                assert t > 0.0

                hit = ro + rd * t
                r = hit.x * hit.x + hit.y * hit.y
                if r > self.aperture_radius[i] * self.aperture_radius[i]:  # out of the element aperture
                    ok = False
                    break

                self.draw_rays[ind, i] = ro
                ro = ti.Vector([hit.x, hit.y, hit.z])

                if not is_stop:
                    # refracted by lens
                    etaI = 1.0 if i == 0 or self.eta[i - 1] == 0.0 else self.eta[i - 1]
                    etaT = self.eta[i] if self.eta[i] != 0.0 else 1.0
                    rd.normalized()
                    has_r, d = refract(rd, n, etaI/etaT)
                    if not has_r:
                        ok = False
                        break
                    rd = ti.Vector([d.x, d.y, d.z])

                elemZ += self.thickness[i]

        self.draw_rays[ind, self._elem_count] = ro
        self.draw_rays[ind, self._elem_count + 1] = ro + rd * 20.0

    def get_ray_points(self):
        return self.draw_rays.to_numpy()



from setuptools import setup, find_packages

print(find_packages())
setup(  name='cameray',
        install_requires=[
            'networkx',
            'taichi',
            'numpy'
        ],
        setup_requires=[],
        python_requires='>=3.6, <=3.9',
        version='0.0.1',

        author='Shuoliu Yang',

        author_email='visysl@outlook.com',

        url='https://github.com/yslib/Cameray',

        package_dir={'':'src'},
        packages=find_packages(where='src'),

        entry_points={
            'console_scripts':[
                'cameray = gui.app:main'
            ]

        },

        classifiers = [
        # https://pypi.org/pypi?%3Aaction=list_classifiers
        'Development Status :: 3 - Alpha',
        'Intended Audience :: End Users/Desktop',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.8',
        ]
        )



import numpy as np
import taichi as ti
from realistic import RealisticCamera
ti.init(arch=ti.gpu)

pos = [10000.0,10000.0,10000.0]   # mm
center = [0.0,0.0,0.0]
world_up = [0.0,1.0,0.0]
resolution = [400, 400]
cam = RealisticCamera(resolution, pos, center, world_up)

pos = [10.0,10.0,10.0 ,1.0]
center = [0.0,0.0,0.0, 1.0]
up = [0.0,1.0, 0.0,1.0]


v = ti.Vector([0.0,0.0,0.0,0.0])
v = ti.Vector([*pos])
matr = ti.Matrix([pos,pos,pos,pos])

@ti.kernel
def test_focal_length():
    print('front z: ', cam.front_z())
    print('lens z: ', cam.rear_z())

    for i in ti.static(range(3)):
        x = 0.1 + i * 0.1
        so = ti.Vector([x, 0.0, 2000.0])
        sd = ti.Vector([0.0, 0.0, -1.0])
        fo = ti.Vector([x, 0.0, cam.rear_z() - 1.0])
        fd = ti.Vector([0.0,0.0,1.0])
        ok1, o1, d1 = cam.gen_ray_from_scene(so, sd)
        ok2, o2, d2 = cam.gen_ray_from_film(fo, fd)
        tf = -o1.x / d1.x
        tf2 = -o2.x / d2.x

    fz1,pz1, fz2, pz2 = cam.compute_thick_lens_approximation()
    print('fz1, pz1, fz2, pz2: ', fz1, pz1, fz2, pz2)
    print('first focal length, second focal length', cam.get_focal_length())

@ti.kernel
def test_pupils():
    cam.refocus(1200)
    cam.recompute_exit_pupil()

# test_pupils()

cam.refocus(10000.0)
cam.recompute_exit_pupil()

@ti.kernel
def gen_ray_test():
    cam.vignet[None] = 0
    for i in range(600):
        for j in range(400):
            weight, ray = cam.gen_ray_of(i, j)
            if weight > 0.0:
                print(i, j)

gen_ray_test()
